unit uCLI;
{$INCLUDE 'ChromePass.inc'}

interface

uses
  SysUtils, Classes,
  uConsoleApp, uChromePass, uConsts;

type
  TChromePassCLI = class(TConsoleApp)
  protected
    function AppTitle: string; override;
    function AppVersion: string; override;
    function DoExecute: integer; override;
    procedure ShowSyntax(const Error: string = ''); override;
    procedure ParamInitialize; override;
  protected
    type TChromePassCommand = (cmNone, cmExport, cmImport);
  private
    fChromePass: TChromePass;
    fCommand: TChromePassCommand;
    fCleanImport: boolean;
    function GetChromePass: TChromePass;
  protected
    function ParseCommandLine: integer; virtual;
    property ChromePass: TChromePass read GetChromePass;
    property Command: TChromePassCommand read fCommand write fCommand;
  public
    constructor Create; override;
    destructor Destroy; override;
  end;

implementation

{ TChromePassCLI }

function TChromePassCLI.AppTitle: string;
begin
  Result := 'ChromePass';
end;

function TChromePassCLI.AppVersion: string;
begin
  Result := {$INCLUDE 'ChromePass.version'} ;
end;

constructor TChromePassCLI.Create;
begin
  inherited;
end;

destructor TChromePassCLI.Destroy;
begin
  FreeAndNil(fChromePass);
  inherited;
end;

function TChromePassCLI.DoExecute: integer;
begin
  // Read command line arguments
  ParamInitialize;
  Result := ParseCommandLine;
  if Result <> EXITCODE_OK then
    Exit
  else
    case Command of
      cmExport:
        Result := ChromePass.ExportPasswords;
      cmImport:
        Result := ChromePass.ImportPasswords(fCleanImport);
      else
        begin
          ShowSyntax('Command not provided.');
          Result := EXITCODE_SYNTAX;
        end;
      end;
end;

function TChromePassCLI.GetChromePass: TChromePass;
begin
  if fChromePass = nil then
    fChromePass := TChromePass.Create;
  Result := fChromePass;
end;

procedure TChromePassCLI.ParamInitialize;
begin
  inherited;
  Command := cmNone;
  ChromePass.ChromeFileName := '';
  ChromePass.DataFileName := '';
  ChromePass.Password := '';
  fCleanImport := False;
end;

function TChromePassCLI.ParseCommandLine: integer;
const
  ONLY_ONE_COMMAND = 'Only one of -e/-i commands may be present.';
  NEED_FILENAME = 'File name must be provided for %s command.';
  NEED_FILENAME_ARGUMENT = 'File name must be provided with %s argument.';
  FILE_NOT_FOUND = 'Input file "%s" does not exist.';
  UNKNOWN_ARGUMENT = 'Unknown argument "%s".';

  procedure SyntaxError(const Msg: string); overload;
    begin
      ShowSyntax(Msg);
      Abort;
    end;
  procedure SyntaxError(const Msg: string; const Params: array of const); overload;
    begin
      SyntaxError(Format(Msg, Params));
    end;

var
  Param: string;
begin
  try
    while ParamGetNext(Param) do
      begin
      if (Param = '-h') or (Param = '-?') then
        SyntaxError('')
      else if Param = '-i' then
        if Command = cmNone then
          if ParamGetNext(Param) then
            if FileExists(Param) then
              begin
              Command := cmImport;
              ChromePass.ChromeFileName := Param;
              end
            else
              SyntaxError(FILE_NOT_FOUND, [Param])
          else
            SyntaxError(NEED_FILENAME, ['-i'])
        else
          SyntaxError(ONLY_ONE_COMMAND)
      else if Param = '-e' then
        if Command = cmNone then
          if ParamGetNext(Param) then
            if FileExists(Param) then
              begin
              Command := cmExport;
              ChromePass.ChromeFileName := Param;
              end
            else
              SyntaxError(FILE_NOT_FOUND, [Param])
          else
            SyntaxError(NEED_FILENAME, ['-e'])
        else
          SyntaxError(ONLY_ONE_COMMAND)
      else if Param = '-f' then
        if ParamGetNext(Param) then
          if Param = '-' then
            ChromePass.DataFileName := ''
          else
            ChromePass.DataFileName := Param
        else
          SyntaxError(NEED_FILENAME_ARGUMENT, ['-f'])
      else if Param = '-pwd' then
        if ParamGetNext(Param) then
          ChromePass.Password := Param
        else
          SyntaxError('Passwords needs to be provided with -pwd')
      else if Param = '-p' then
        begin
        Write('Enter password: ');
        ChromePass.Password := ReadPassword;
        Writeln;
        end
      else if Param = '-clean' then
        fCleanImport := True
      else
        SyntaxError(UNKNOWN_ARGUMENT, [Param])
      end;
    Result := EXITCODE_OK;
  except
    on EAbort do
      Result := EXITCODE_SYNTAX;
    end;
end;

procedure TChromePassCLI.ShowSyntax(const Error: string);
begin
  inherited;
  WriteColored(ccWhite, '[<arg>] [<arg>] ... '); Writeln;
  Writeln;
  Writeln('Commands:');
  WriteColored(ccWhite, ' -h, -?'); Writeln(' .......... Show this help screen.');
  WriteColored(ccWhite, ' -e <infile>'); Writeln(' ..... Exports passwords from Chrome''s <infile>.');
  WriteColored(ccWhite, ' -i <infile>'); Writeln(' ..... Imports passwords from text <infile>.');
  Writeln;
  Writeln('Arguments:');
  WriteColored(ccWhite, ' -f <filename>'); Writeln(' ... Resulting data will be written to file <file>.');
  WriteColored(ccWhite, ' -p'); Writeln(' .............. Encrypt passwords using a manually provided password.');
  WriteColored(ccWhite, ' -pwd <pass>'); Writeln(' ..... Encrypt passwords using a password <pass>.');
  WriteColored(ccWhite, ' -clean'); Writeln(' .......... When importing, erase previously stored logins.');
end;

end.
