unit uChromePass;
{$INCLUDE 'ChromePass.inc'}

interface

uses
  SysUtils, Classes, Windows,
  SqliteTable3,
  NativeXml,
  SynaCode,
  DCPrijndael, DCPsha256, DCPcrypt2,
  uXml, uConsts;

type
  TChromePass = class
  private
    fInputFileName: string;
    fOutputFileName: string;
    fSalt: string;
    fPassword: string;
    fIterationCount: integer;
    fHashObj: TDCP_hash;
    fCryptObj: TDCP_blockcipher;
    function GetCryptObj: TDCP_blockcipher;
    function GetHashObj: TDCP_hash;
    procedure SetIterationCount(const Value: integer);
    procedure SetPassword(const Value: string);
    procedure SetSalt(const Value: string);
  protected
    {$IFDEF UNICODE}
    function Hash(const Data: string): string; overload;
    {$ENDIF}
    function Hash(const Data: RawByteString): RawByteString; overload;
    procedure Encrypt(var Data; DataSize: integer);
    procedure Decrypt(var Data; DataSize: integer);
    procedure XmlSetString(Node: TXmlNode; const Name, Data: string);
    procedure XmlSetInteger(Node: TXmlNode; const Name: string; Data: integer);
    procedure XmlSetBinary(Node: TXmlNode; const Name: string; const Data; DataSize: integer);
    procedure XmlSetField(Node: TXmlNode; const Name: string; Table: TSqliteTable; Row: integer);
    procedure SqlSetField(var Cols, Vals: string; Node: TXmlNode; const Name: string; DataType: integer; Row: integer);
    function IsRequiredField(const Name: string): boolean;
    function IsPasswordField(const Name: string): boolean;
    procedure InitializeSalt;
    property HashObj: TDCP_hash read GetHashObj;
    property CryptObj: TDCP_blockcipher read GetCryptObj;
    property IterationCount: integer read fIterationCount write SetIterationCount;
    property Salt: string read fSalt write SetSalt;
  public
    constructor Create;
    destructor Destroy; override;
    function ExportPasswords: integer;
    function ImportPasswords(Clean: boolean): integer;
    property ChromeFileName: string read fInputFileName write fInputFileName;
    property DataFileName: string read fOutputFileName write fOutputFileName;
    property Password: string read fPassword write SetPassword;
  end;

implementation

const
  PasswordField = 'password_value';
  RequiredFields: array[0..2] of string = (
    'origin_url', 'username_value', PasswordField
  );
  FixedSalt = 'f2D|9@t9oeI,-Mk@4`sJD"\c.4tCxRL+';
  FixedPassword = '5XrN>3gaPOozU_T3cmp!:9x\e|Uo2D{~';
  MaxColumnsSupported = 1000;
  DefaultSaltLength = 64;

type
  TDataBlob = record
    cbData: DWORD;
    pbData: HLOCAL;
  end;
  PDataBlob = ^TDataBlob;

const
  MS_ENHANCED_PROV = 'Microsoft Enhanced Cryptographic Provider v1.0';
  PROV_RSA_FULL = 1;
  CRYPT_VERIFYCONTEXT = $F0000000;
  CRYPT_NEWKEYSET = 8;

function CryptUnprotectData(const DataIn: TDataBlob; Description: PWideChar; Entropy: PDataBlob; Reserved: Pointer; PromptStruct: Pointer; Flags: DWORD; out DataOut: TDataBlob): BOOL; stdcall; external 'crypt32.dll' name 'CryptUnprotectData';
function CryptProtectData(const DataIn: TDataBlob; Description: PWideChar; Entropy: PDataBlob; Reserved: Pointer; PromptStruct: Pointer; Flags: DWORD; out DataOut: TDataBlob): BOOL; stdcall; external 'crypt32.dll' name 'CryptProtectData';
function CryptAcquireContext(var Handle: THandle; Container, Provider: PChar; ProviderType, Flags: DWORD): BOOL; stdcall; external 'advapi32.dll' name {$IFDEF UNICODE} 'CryptAcquireContextW' {$ELSE} 'CryptAcquireContextA' {$ENDIF} ;
function CryptGenRandom(Handle: THandle; Size: DWORD; var Buffer): BOOL; stdcall; external 'advapi32.dll' name 'CryptGenRandom';
function CryptReleaseContext(Handle: THandle; Flags: DWORD): BOOL; stdcall; external 'advapi32.dll' name 'CryptReleaseContext';

{ TChromePass }

constructor TChromePass.Create;
begin
  inherited;
  fIterationCount := 10000;
  InitializeSalt;
end;

procedure TChromePass.Decrypt(var Data; DataSize: integer);
var
  Buffer: RawByteString;
begin
  if DataSize > 0 then
    begin
    SetLength(Buffer, DataSize);
    CryptObj.DecryptCTR(Data, Buffer[1], DataSize);
      // Note: This function uses a fixed starting counter value, which is generally
      // a poor practice, but in this case is actually useful: We want to be able
      // to copy encrypted passwords to new records.
    Move(Buffer[1], Data, DataSize);
    end;
end;

destructor TChromePass.Destroy;
begin
  FreeAndNil(fHashObj);
  FreeAndNil(fCryptObj);
  inherited;
end;

procedure TChromePass.Encrypt(var Data; DataSize: integer);
var
  Buffer: RawByteString;
begin
  if DataSize > 0 then
    begin
    SetLength(Buffer, DataSize);
    CryptObj.EncryptCTR(Data, Buffer[1], DataSize);
      // Note: This function uses a fixed starting counter value, which is generally
      // a poor practice, but in this case is actually useful: We want to be able
      // to copy encrypted passwords to new records.
    Move(Buffer[1], Data, DataSize);
    end;
end;

function TChromePass.ExportPasswords: integer;
var
  DB: TSqliteDatabase;
  Table: TSQLiteTable;
  Xml: TXmlDoc;
  OptionsNode, PasswordsNode, PasswordNode: TXmlNode;
  i: Integer;
begin
  if ChromeFileName = '' then
    Raise EInOutError.Create('Chrome file must not be empty.');
  DB := TSqliteDatabase.Create(ChromeFileName);
  try
    Xml := TXmlDoc.Create;
    try
      Xml.Root.Name := 'chromepass';
      OptionsNode := Xml.Root.NodeNew('options');
      if Password <> '' then
        begin
        InitializeSalt;
        XmlSetString(OptionsNode, 'salt', Salt);
        XmlSetInteger(OptionsNode, 'iteration_count', IterationCount);
        XmlSetString(OptionsNode, 'password', Hash(Salt + Password + FixedSalt));
        end;
      PasswordsNode := Xml.Root.NodeNew('passwords');
      Table := DB.GetTable('SELECT * FROM logins ORDER BY origin_url, username_value, date_created');
      try
        while not Table.Eof do
          begin
          PasswordNode := PasswordsNode.NodeNew('item');
          for i := Low(RequiredFields) to High(RequiredFields) do
            XmlSetField(PasswordNode, RequiredFields[i], Table, Table.Row);
          for i := 0 to Pred(Table.ColCount) do
            if not IsRequiredField(Table.Columns[i]) then
              XmlSetField(PasswordNode, Table.Columns[i], Table, Table.Row);
          Table.Next;
          end;
        Xml.SetIndentation('  ');
        if DataFileName <> '' then
          Xml.SaveToFile(DataFileName)
        else
          begin
          Writeln(Xml.SaveToBinaryString);
          end;
        Result := EXITCODE_OK;
      finally
        FreeAndNil(Table);
      end;
    finally
      FreeAndNil(Xml);
    end;
  finally
    FreeAndNil(DB);
  end;
end;

function TChromePass.GetCryptObj: TDCP_blockcipher;
var
  Key1, Key2: RawByteString;
  i, n: integer;
begin
  if fCryptObj = nil then
    begin
    Key1 := {$IFDEF UNICODE} RawByteString {$ENDIF} (Salt + Password + FixedPassword);
    n := IterationCount;
    if n < 100 then
      n := 100;
    for i := 0 to Pred(n) do
      begin
      Key2 := Hash(Key1 + {$IFDEF UNICODE} RawByteString {$ENDIF} (IntToStr(i)));
      ZeroMemory(@Key1[1], Length(Key1)); {$MESSAGE HINT 'SecureZeroMemory would be better'}
      Key1 := Key2;
      ZeroMemory(@Key2[1], Length(Key2));
      end;
    fCryptObj := TDCP_rijndael.Create(nil);
    fCryptObj.Init(Key1[1], Length(Key1), nil); // no IV needed or desired
    ZeroMemory(@Key1[1], Length(Key1));
    end;
  Result := fCryptObj;
end;

function TChromePass.GetHashObj: TDCP_hash;
begin
  if fHashObj = nil then
    fHashObj := TDCP_sha256.Create(nil);
  Result := fHashObj;
end;

function TChromePass.Hash(const Data: RawByteString): RawByteString;
var
  Digest: RawByteString;
begin
  HashObj.Init;
  HashObj.UpdateStr(SynaCode.EncodeBase64(Data));
  SetLength(Digest, HashObj.GetHashSize div 8);
  HashObj.Final(Digest[1]);
  HashObj.Burn;
  Result := SynaCode.EncodeBase64(Digest);
end;

{$IFDEF UNICODE}
function TChromePass.Hash(const Data: string): string;
begin
  Result := string(Hash(RawByteString(Data)));
end;
{$ENDIF}

function TChromePass.ImportPasswords(Clean: boolean): integer;
type
  TColumnInfo = record
    Name: string;
    DataType: integer;
    end;
var
  DB: TSqliteDatabase;
  Table: TSQLiteTable;
  Xml: TXmlDoc;
  OptionsNode, PasswordsNode, PasswordNode: TXmlNode;
  s, PasswordHash, ColumnList, ValueList: string;
  i, j, n, Row: Integer;
  Columns: array of TColumnInfo;
begin
  if DataFileName = '' then
    Raise EInOutError.Create('Data file must not be empty.');
  if ChromeFileName = '' then
    Raise EInOutError.Create('Chrome file must not be empty.');
  Result := EXITCODE_NO_DATA;
  DB := TSqliteDatabase.Create(ChromeFileName);
  try
    Xml := TXmlDoc.Create;
    try
      Xml.LoadFromFile(DataFileName);
      if XmlNodeByPath(Xml, 'options', OptionsNode) then
        if XmlAttribute(OptionsNode, 'password', PasswordHash) then
          if PasswordHash <> '' then
            begin
            if XmlAttribute(OptionsNode, 'iteration_count', s) then
              IterationCount := StrToInt(s);
            if XmlAttribute(OptionsNode, 'salt', s) then
              Salt := s;
            if PasswordHash <> Hash(Salt + Password + FixedSalt) then
              Raise Exception.CreateFmt('Invalid password for file %s', [DataFileName]);
            end;
      if XmlNodeByPath(Xml, 'passwords', PasswordsNode) then
        begin
        DB.BeginTransaction;
        try
          if Clean then
            DB.ExecSQL('DELETE FROM logins');
          n := 0;
          SetLength(Columns, MaxColumnsSupported);
          Table := DB.GetTable('PRAGMA TABLE_INFO(logins)');
          try
            while not Table.EOF do
              begin
              Columns[n].Name := Table.FieldByName['name'];
              s := Table.FieldByName['type'];
              if AnsiCompareText(s, 'BLOB') = 0 then
                Columns[n].DataType := dtBlob
              else if AnsiCompareText(s, 'INTEGER') = 0 then
                Columns[n].DataType := dtInt
              else if AnsiCompareText(s, 'REAL') = 0 then
                Columns[n].DataType := dtNumeric
              else if AnsiCompareText(s, 'NULL') = 0 then
                Columns[n].DataType := dtNull
              else
                Columns[n].DataType := dtStr;
              Inc(n);
              Table.Next;
              end;
          finally
            FreeAndNil(Table);
            end;
          SetLength(Columns, n);
          Row := 0;
          for i := 0 to Pred(PasswordsNode.NodeCount) do
            begin
            PasswordNode := PasswordsNode.Nodes[i];
            if PasswordNode.ElementType = xeNormal then
              if PasswordNode.Name = 'item' then
                begin
                ColumnList := '';
                ValueList := '';
                for j := 0 to High(Columns) do
                  SqlSetField(ColumnList, ValueList, PasswordNode, Columns[j].Name, Columns[j].DataType, Row);
                if ColumnList <> '' then
                  begin
                  s := 'INSERT INTO logins (' + ColumnList + ') VALUES (' + ValueList + ')';
                  //Writeln(s);
                  DB.ExecSQL(s);
                  end;
                Inc(Row);
                end;
            end;
          DB.Commit;
          Result := EXITCODE_OK;
        except
          DB.Rollback;
          Raise;
          end;
        end;
    finally
      FreeAndNil(Xml);
      end;
  finally
    FreeAndNil(DB);
  end;
end;

procedure TChromePass.InitializeSalt;
var
  ContextValid: boolean;
  LastError: DWORD;
  Handle: THandle;
  BinarySalt: RawByteString;
begin
  FreeAndNil(fCryptObj);
  LastError := NO_ERROR;
  ContextValid := False;
  if CryptAcquireContext(Handle, nil, MS_ENHANCED_PROV, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT) then
    ContextValid := True
  else
    begin
    LastError := GetLastError;
    if LastError = DWORD(NTE_BAD_KEYSET) then
      if CryptAcquireContext(Handle, nil, MS_ENHANCED_PROV, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT or CRYPT_NEWKEYSET) then
        ContextValid := True
      else
        LastError := GetLastError;
    end;
  if not ContextValid then
    RaiseLastOSError(LastError)
  else
    try
      SetLength(BinarySalt, DefaultSaltLength);
      if CryptGenRandom(Handle, DefaultSaltLength, BinarySalt[1]) then
        fSalt := {$IFDEF UNICODE} string {$ENDIF} (SynaCode.EncodeBase64(BinarySalt))
      else
        RaiseLastOSError;
    finally
      CryptReleaseContext(Handle, 0);
    end;
end;

function TChromePass.IsPasswordField(const Name: string): boolean;
begin
  Result := AnsiCompareText(Name, PasswordField) = 0;
end;

function TChromePass.IsRequiredField(const Name: string): boolean;
var
  i: integer;
begin
  Result := False;
  for i := Low(RequiredFields) to High(RequiredFields) do
    if AnsiCompareText(Name, RequiredFields[i]) = 0 then
      begin
      Result := True;
      Break;
      end;
end;

procedure TChromePass.SetIterationCount(const Value: integer);
begin
  fIterationCount := Value;
  FreeAndNil(fCryptObj);
end;

procedure TChromePass.SetPassword(const Value: string);
begin
  fPassword := Value;
  FreeAndNil(fCryptObj);
end;

procedure TChromePass.SetSalt(const Value: string);
begin
  fSalt := Value;
  FreeAndNil(fCryptObj);
end;

procedure TChromePass.SqlSetField(var Cols, Vals: string; Node: TXmlNode; const Name: string; DataType: integer; Row: integer);
var
  InData, OutData: TDataBlob;
  Value: string;
  BlobValue: RawByteString;
begin
  if Node.AttributeIndexByname( {$IFDEF UNICODE} UTF8String {$ENDIF} (Name)) < 0 then
    Value := 'NULL'
  else
    begin
    Value := XmlAttribute(Node, Name);
    if DataType = dtBlob then
      begin
      BlobValue := SynaCode.DecodeBase64( {$IFDEF UNICODE} RawByteString {$ENDIF} (Value));
      if BlobValue = '' then
        Value := 'X'''''
      else
        begin
        if IsPasswordField(Name) then
          begin
          InData.cbData := Length(BlobValue);
          InData.pbData := HLOCAL(@BlobValue[1]);
          if Password <> '' then
            Decrypt(Pointer(InData.pbData)^, InData.cbData);
          if CryptProtectData(InData, nil, nil, nil, nil, 1 {CRYPTPROTECT_UI_FORBIDDEN}, OutData) then
            try
              SetLength(BlobValue, OutData.cbData);
              if OutData.cbData > 0 then
                Move(Pointer(OutData.pbData)^, BlobValue[1], OutData.cbData);
            finally
              LocalFree(OutData.pbData);
            end
          else
            Raise Exception.CreateFmt('Failed to encrypt password, item=%d, col=%s', [Row, Name]);
          end;
        Value := 'X''' + {$IFDEF UNICODE} string {$ENDIF} (EncodeBinHex(BlobValue)) + '''';
        end;
      end
    else
      begin
      Value := AnsiQuotedStr(Value, '''');
      end;
    end;
  if Cols = '' then
    begin
    Cols := Name;
    Vals := Value;
    end
  else
    begin
    Cols := Cols + ',' + Name;
    Vals := Vals + ',' + Value;
    end;
end;

procedure TChromePass.XmlSetBinary(Node: TXmlNode; const Name: string; const Data; DataSize: integer);
var
  DataStr: RawByteString;
begin
  SetLength(DataStr, DataSize);
  if DataSize > 0 then
    begin
    Move(Data, DataStr[1], DataSize);
    DataStr := SynaCode.EncodeBase64(DataStr);
    end;
  XmlSetString(Node, Name, {$IFDEF UNICODE} string {$ENDIF} (DataStr));
end;

procedure TChromePass.XmlSetField(Node: TXmlNode; const Name: string; Table: TSqliteTable; Row: integer);
var
  Index: integer;
  FieldData: TMemoryStream;
  InData, OutData: TDataBlob;
begin
  Index := Table.FieldIndex[Name];
  if not Table.FieldIsNull(Index) then
    if Table.ColumnTypes[Index] = dtBlob then
      begin
      FieldData := Table.FieldAsBlob(Index);
      if FieldData.Size > 0 then
        if IsPasswordField(Name) then
          begin
          InData.cbData := FieldData.Size;
          InData.pbData := HLOCAL(FieldData.Memory);
          if CryptUnprotectData(InData, nil, nil, nil, nil, 1 {CRYPTPROTECT_UI_FORBIDDEN}, OutData) then
            try
              if Password <> '' then
                Encrypt(Pointer(OutData.pbData)^, OutData.cbData);
              XmlSetBinary(Node, Name, Pointer(OutData.pbData)^, OutData.cbData);
            finally
              LocalFree(OutData.pbData);
            end
          else
            Raise Exception.CreateFmt('Failed to decrypt password, row=%d, col=%s', [Row, Name]);
          end
        else
          XmlSetBinary(Node, Name, FieldData.Memory^, FieldData.Size);
      end
    else
      XmlSetString(Node, Name, Table.FieldAsString(Index));
end;

procedure TChromePass.XmlSetInteger(Node: TXmlNode; const Name: string; Data: integer);
begin
  XmlSetString(Node, Name, IntToStr(Data));
end;

procedure TChromePass.XmlSetString(Node: TXmlNode; const Name, Data: string);
begin
  Node.AttributeByNameWide[{$IFDEF UNICODE} UTF8String {$ENDIF} (Name)] := Data;
end;

end.
